import socket
from ipaddress import ip_address, AddressValueError

def resolve_domain_or_ip(url):
  """Resolves a domain name to its IP address or verifies an IP address.

  Parameters
  ----------
  url : str
      The domain name or IP address.

  Returns
  -------
  tuple: (bool, str, str)
      <success>, <ip>, <original_url>
      A tuple containing a boolean indicating if the resolution was successful,
      a string with the resolved or verified IP address, and the original URL.
  """
  try:
    # Check if the URL is an IP address
    ip_address(url)
    return True, url, url
  except:
    pass

  try:
    # Try to resolve the domain name to an IP address
    ip_of_url = socket.gethostbyname(url)
    return True, ip_of_url, url
  except:
    return False, None, url
  
  
  

if __name__ == '__main__':
  # Usage:
  url = "r9092118.ala.eu-central-1.emqxsl.com"
  success, ip, original_url = resolve_domain_or_ip(url)
  if not success:
    print(f"Cannot connect to {original_url} (IP: {ip})")
  else:
    print(f"Resolved {original_url} to IP: {ip}")
    


"""
ex01_part1_connect.py
---------------------
This is a simple example of how to use the ratio1 SDK.

In this example, we connect to the network, listen for heartbeats from 
  ratio1 Edge Protocol edge nodes and print the CPU of each node.
  
  
  
New connection algorithm:

1. The client connects to the network.
2. The client waits for the first 2 supervisors payload with network map. 
3. The client reads `partner_peers` - all nodes that recognize the client as allowed - based on supervisor(s) network map.
4. The client sends a pipeline status to all `partner_peers`.
5. The client then knows the `partner_peers` and can send messages to them.
IMPORTANT: Session will WAIT until network map is clarified.

"""
import json

from ratio1 import Session, Payload, PAYLOAD_DATA, HEARTBEAT_DATA


class MessageHandler:  
  def shorten_address(self, address):
    """
    This method is used to shorten the address of the edge node.
    """
    return address[:8] + "..." + address[-6:]
  
  def on_heartbeat(self, session: Session, node_addr: str, heartbeat: dict):
    """
    This method is called when a heartbeat is received from an edge node.
    
    Parameters
    ----------
    session : Session
        The session object that received the heartbeat.
        
    node_addr : str
        The address of the edge node that sent the heartbeat.
        
    heartbeat : dict
        The heartbeat received from the edge node.        
    """
    node_alias = heartbeat[PAYLOAD_DATA.EE_ID]
    short_addr = self.shorten_address(node_addr)
    cpu = heartbeat[HEARTBEAT_DATA.CPU]
    cpu_load = heartbeat[HEARTBEAT_DATA.CPU_USED]
    mem = heartbeat[HEARTBEAT_DATA.MACHINE_MEMORY]
    mem_avail = heartbeat[HEARTBEAT_DATA.AVAILABLE_MEMORY]
    disk_avail = heartbeat[HEARTBEAT_DATA.AVAILABLE_DISK]
    disk_total = heartbeat[HEARTBEAT_DATA.TOTAL_DISK]
    disk_load = (disk_total - disk_avail) / disk_total * 100
    mem_load = (mem - mem_avail) / mem * 100
    self.hb = heartbeat # simple storage of the last heartbeat
    plugins = heartbeat[HEARTBEAT_DATA.ACTIVE_PLUGINS] or []
    pipelines = heartbeat[HEARTBEAT_DATA.CONFIG_STREAMS] or []
    session.P(
      "<{}> {} {} Load: {:.1f}% CPU, {:.1f}% RAM, {:.1f}% Disk, {} plugins, {} pipelines".format(
        short_addr, node_alias, cpu, cpu_load, mem_load, disk_load, len(plugins), len(pipelines)
      ),
      color='magenta'
    )
    return



if __name__ == '__main__':
  # create a naive message handler   
  filterer = MessageHandler()
  
  # create a session
  # the network credentials are read from the .env file automatically
  session = Session(
      on_heartbeat=filterer.on_heartbeat,
      verbosity=3
  )

  session.P("Client address is: {}".format(session.get_client_address()), color='m')
  
  # lets see top 5 online nodes
  netinfo = session.get_network_known_nodes(online_only=True)
  session.P(f"Online nodes reported by {netinfo.reporter}:\n{netinfo.report}")

  session.sleep(10) # wait for 10 seconds
  session.P("Closing session...", color='m')
  session.close()
  session.P("Main thread exiting...", color='m')
  print(json.dumps(filterer.hb, indent=2))



import logging
import os
import time
import datetime
import calendar
import warnings


try:
  import zoneinfo  # pragma: no cover
except ImportError:
  from backports import zoneinfo  # pragma: no cover

from . import windows_tz

log = logging.getLogger("tzlocal")


def get_tz_offset(tz):
  """Get timezone's offset using built-in function datetime.utcoffset()."""
  return int(datetime.datetime.now(tz).utcoffset().total_seconds())


def assert_tz_offset(tz, error=True):
  """Assert that system's timezone offset equals to the timezone offset found.

  If they don't match, we probably have a misconfiguration, for example, an
  incorrect timezone set in /etc/timezone file in systemd distributions.

  If error is True, this method will raise a ValueError, otherwise it will
  emit a warning.
  """

  tz_offset = get_tz_offset(tz)
  system_offset = calendar.timegm(time.localtime()) - calendar.timegm(time.gmtime())
  # No one has timezone offsets less than a minute, so this should be close enough:
  if abs(tz_offset - system_offset) > 60:
    msg = (
        "Timezone offset does not match system offset: {} != {}. "
        "Please, check your config files."
    ).format(tz_offset, system_offset)
    if error:
      raise ValueError(msg)
    warnings.warn(msg)


def _tz_name_from_env(tzenv=None):
  if tzenv is None:
    tzenv = os.environ.get("TZ")

  if not tzenv:
    return None

  log.debug(f"Found a TZ environment: {tzenv}")

  if tzenv[0] == ":":
    tzenv = tzenv[1:]

  if tzenv in windows_tz.tz_win:
    # Yup, it's a timezone
    return tzenv

  if os.path.isabs(tzenv) and os.path.exists(tzenv):
    # It's a file specification, expand it, if possible
    parts = os.path.realpath(tzenv).split(os.sep)

    # Is it a zone info zone?
    possible_tz = "/".join(parts[-2:])
    if possible_tz in windows_tz.tz_win:
      # Yup, it is
      return possible_tz

    # Maybe it's a short one, like UTC?
    if parts[-1] in windows_tz.tz_win:
      # Indeed
      return parts[-1]

  log.debug("TZ does not contain a time zone name")
  return None


def _tz_from_env(tzenv=None):
  if tzenv is None:
    tzenv = os.environ.get("TZ")

  if not tzenv:
    return None

  # Some weird format that exists:
  if tzenv[0] == ":":
    tzenv = tzenv[1:]

  # TZ specifies a file
  if os.path.isabs(tzenv) and os.path.exists(tzenv):
    # Try to see if we can figure out the name
    tzname = _tz_name_from_env(tzenv)
    if not tzname:
      # Nope, not a standard timezone name, just take the filename
      tzname = tzenv.split(os.sep)[-1]
    with open(tzenv, "rb") as tzfile:
      return zoneinfo.ZoneInfo.from_file(tzfile, key=tzname)

  # TZ must specify a zoneinfo zone.
  try:
    tz = zoneinfo.ZoneInfo(tzenv)
    # That worked, so we return this:
    return tz
  except zoneinfo.ZoneInfoNotFoundError:
    # Nope, it's something like "PST4DST" etc, we can't handle that.
    raise zoneinfo.ZoneInfoNotFoundError(
        "tzlocal() does not support non-zoneinfo timezones like %s. \n"
        "Please use a timezone in the form of Continent/City" % tzenv
    ) from None



from ratio1 import Logger

if __name__ == "__main__":
  l = Logger("DT")
  
  UTC_DATE = "2025-02-01 12:00:00"
  converted = l.utc_to_local(
    remote_datetime=UTC_DATE, 
    remote_utc=None, 
    as_string=False
  )
  l.P(f"UTC {UTC_DATE} to local {converted}")


import time
from .pipeline import Pipeline



class WebappPipeline(Pipeline):
  
  def __init__(
    self, 
    session, 
    log, 
    *, 
    node_addr, 
    name, 
    config={}, 
    plugins=[], 
    is_attached=False, 
    on_data=None,
    on_notification=None,
    existing_config=None, 
    extra_debug=False,
    **kwargs
  ) -> None:
    """
    This is a special type of pipeline that is used to deploy webapps.
    It will override the deploy method to return the ngrok URL as well as the on_data method to extract the ngrok URL.
    """
    self.ngrok_url = None
    self.__extra_debug = extra_debug
    _on_data = [self.__check_payloads, on_data]    
    super().__init__(
      session, 
      log, 
      node_addr=node_addr, 
      name=name, 
      config=config, 
      plugins=plugins, 
      is_attached=is_attached, 
      on_data=_on_data,
      on_notification=on_notification,
      existing_config=existing_config, 
      
      **kwargs
    )
    return
  
  
  def __check_payloads(self, session, plugin_signature, plugin_instance, data):
    """
    Check if the payload if from ngrok and extract the ngrok URL.
    
    Parameters
    ----------
    
    
    session : Session
        The session object that received the payload.
            
    plugin_signature : str
        The signature of the plugin that sent the payload.
        
    plugin_instance : str
        The instance of the plugin that sent the payload.
        
    data : Payload
        The payload received from the edge node.     
        
         
    """
    if self.__extra_debug:
      self.P(f"Received payload from {plugin_signature} ({plugin_instance})")
    if "NGROK_URL" in data:
        self.ngrok_url = data["NGROK_URL"]
    return
    
  def create_plugin_instance(
    self, 
    *, 
    signature, 
    instance_id, 
    config={}, 
    ngrok_edge_label=None,
    **kwargs
  ):

    if ngrok_edge_label is not None:
      self.ngrok_url = "URL_DEFINED_IN_EDGE_LABEL"
      
    return super().create_plugin_instance(
      signature=signature, 
      instance_id=instance_id, 
      config=config, 
      ngrok_edge_label=ngrok_edge_label,
      **kwargs
    )    
  
  def deploy(self, verbose=False, timeout=15, **kwargs):
    """
    Deploy the pipeline and return the URL of the webapp.
    
    Returns
    -------
    str
        The URL of the webapp.
    """
    res = super().deploy(verbose=verbose, timeout=timeout, **kwargs)
    # now we wait for the ngrok url to be available
    start = time.time()
    while self.ngrok_url is None:
      elapsed = time.time() - start
      if elapsed > timeout:
        msg = "Timeout waiting for ngrok url"
        self.P(msg, color="red")
        raise Exception(msg)
    # return the ngrok url
    return self.ngrok_url


import os
import json


from ratio1 import Logger
from ratio1.bc import DefaultBlockEngine



if __name__ == '__main__' :
  
  NETWORKS = [
    "mainnet",
    "testnet",
    "devnet",
  ]
  
  network = NETWORKS[2]

  l = Logger("ENC")
  eng = DefaultBlockEngine(
    log=l, name="default", 
    user_config=True
  )
  
  eng.reset_network(network)

  oracles = eng.web3_get_oracles(debug=True)
  l.P("\nOracles for {}:\n {}".format(
      network, json.dumps(oracles, indent=2)
    ), 
    color='b', show=True
  )
  
  eng.reset_network("devnet")
  
  NODE = "0xED5BE902866855caabF1Acb558209FC40E62524A"
  
  info = eng.web3_get_node_info(node_address=NODE)
  l.P(f"Node info for {NODE}:\n{json.dumps(info, indent=2)}", color='m')
  
  WALLET = "0x464579c1Dc584361e63548d2024c2db4463EdE48"
  
  nodes = eng.web3_get_wallet_nodes(address=WALLET)
  l.P(f"Nodes for {WALLET}:")
  for n in nodes:
    if n in oracles:
      l.P(f"  {n} (oracle)", color='g')
    else:
      l.P(f"  {n}")
  
  
  



"""
ex16_launch_repo_based_http_server.py
-------------------------------

This script demonstrates how to create and deploy a http server based on a public repository using ngrok
edge definition. The script will create a webapp based on a public repository containing some static assets
and deploy it to the target Ratio1 Edge Node and assign it the same ngrok edge label.

TODO:
This will be modified so that it will start a generic http server that will generate a dynamic URL which can
be used to access the webapp. For this demo, we will use a custom webapp that resides at a
public repository location.

"""
import os
from ratio1 import Session


if __name__ == "__main__":
  # 1. Create the Ratio1 Session object.
  sess = Session()
  
  # 2. Wait for the target node to be seen as active.
  nodes = [
    os.environ.get("EE_TARGET_NODE_1")
  ]
  for node in nodes:
    sess.wait_for_node(node)
  # endfor node in nodes

  # 3. Retrieve the Ngrok edge label. (optional)
  # In case this is not provided and create_and_deploy_balanced_http_server is used an exception will be raised.
  # In case this is not provided by the user, but create_http_server method is used,
  # a unique URL will be generated by the ngrok service.
  # Warning! create_http_server method does not support multiple nodes.
  ngrok_edge_label = os.environ.get("NGROK_EDGE_LABEL", None)

  # 4. Define the assets of the web application
  git_username = os.environ.get("GITHUB_USER")
  git_token = os.environ.get("GITHUB_TOKEN")
  assets = {
    # This will specify that the target node(s) need(s) to download
    # the assets from the provided repository.
    # It will also periodically check for updates.
    # "operation": "release_asset",

    # This is the asset that needs to be downloaded.
    # If this is a zip it will be unzipped on the target node(s).
    "asset_filter": "asset1_from_my_release.zip",
    # The repository from which the asset will be downloaded.
    "url": "<github_repo_url>",
    # The following 2 fields are necessary only if the repository is private.
    "username": git_username,
    "token": git_token
  }

  # 5. Define the environment variables for the http server if needed.
  env_vars = {
    # "KEY1": "VALUE1",
  }

  # This is the directory where the assets will be searched for.
  # If, for example, index.html is in <zip_root>/dist/index.html,
  # the static_directory should be "dist".
  # If this is not provided, the root of the zip will be used.
  static_directory = "dist"

  # This will be used instead of a not found error in case the user accesses a non-existing route.
  # If this is not provided, a not found error will be returned.
  default_route = "/"

  # 6. Define endpoints if necessary. This will be used for custom routing and is optional.
  # The below example is the default approach(the index.html will be served at the root).
  endpoints = [
    # {
    #   "endpoint_type": "html",
    #   "web_app_file_name": "index.html",
    #   "endpoint_route": "/",
    # }
  ]

  # 7. Create and deploy the http server.
  # This will create a webapp based on the provided assets and deploy it to the target node(s).
  sess.create_and_deploy_balanced_http_server(
    nodes=nodes,
    name="Ratio1 HTTP Server Deploy Tutorial",
    env_vars=env_vars,  # optional
    assets=assets,
    static_directory=static_directory,
    default_route=default_route,  # optional
    endpoints=endpoints,  # optional
    ngrok_edge_label=ngrok_edge_label,
  )

  # Observations:
  #   Next code is not mandatory - it is used to keep the session open and cleanup the resources.
  #   In production, you would not need this code as the script can close after the pipeline will be sent
  sess.run(
    wait=True,  # wait for the user to stop the execution
    close_pipelines=True  # when the user stops the execution, the remote edge-node pipelines will be closed
  )






from ratio1.base.responses import Response
# from .responses import Response
from time import time, sleep


class Transaction():
  def __init__(self, log, session_id: str, *, lst_required_responses: list[Response] = None, timeout: int = 0, on_success_callback: callable = None, on_failure_callback: callable = None) -> None:
    self.log = log
    self.session_id = session_id
    self.lst_required_responses = lst_required_responses
    self.timeout = timeout

    self.on_success_callback = on_success_callback
    self.on_failure_callback = on_failure_callback

    self.resolved_callback = None
    self.__is_solved = False
    self.__is_finished = False

    self.start_time = time()
    for response in self.lst_required_responses:
      response.set_logger(log)

    return

  def get_unsolved_responses(self) -> list[Response]:
    """
    Returns a list of unsolved responses in the transaction.

    Returns
    -------
    list[Response]
        List of unsolved responses in the transaction.
    """
    return [response for response in self.lst_required_responses if not response.is_solved()]

  def is_solved(self) -> bool:
    """
    Returns whether the transaction is solved or not.
    If the transaction is solved in this call, the resolved_callback is set to the appropriate callback function.

    Returns
    -------
    bool
        Whether the transaction is solved or not.
    """

    # if the transaction is already solved, return True
    if self.__is_solved:
      return True

    # first check if all responses are solved
    all_responses_solved = all([response.is_solved() for response in self.lst_required_responses])
    if all_responses_solved:
      self.__is_solved = True

      all_responses_good = all([response.is_good_response() for response in self.lst_required_responses])
      if all_responses_good:
        self.resolved_callback = self.on_success_callback
      else:
        if not self.on_failure_callback:
          self.resolved_callback = lambda: self.log.P("Transaction failed")
        else:
          fail_reasons = [
            response.fail_reason for response in self.lst_required_responses if not response.is_good_response()]
          self.resolved_callback = lambda: self.on_failure_callback(fail_reasons)
        # endif computing resolved_callback for failure
      # endif computing resolved_callback
      return True
    # endif all responses solved

    # if the transaction is not solved, check if the transaction has timed out
    elapsed_time = time() - self.start_time
    if self.timeout > 0 and elapsed_time > self.timeout:
      # Timeout occurred
      self.__is_solved = True

      fail_reason = f"Transaction timeout ({self.timeout}s). Responses not received: "
      fail_reason += ", ".join([str(response) for response in self.lst_required_responses if not response.is_solved()])

      if not self.on_failure_callback:
        self.resolved_callback = lambda: self.log.P(fail_reason)
      else:
        self.resolved_callback = lambda: self.on_failure_callback([fail_reason])
      return True

    return False

  def is_finished(self) -> bool:
    """
    Returns whether the transaction is finished or not.
    A transaction finishes when the callback is called.

    Returns
    -------
    bool
        Whether the transaction is finished or not.
    """
    return self.__is_finished

  def handle_payload(self, payload: dict) -> None:
    """
    This method is called when a payload is received from the server.
    This method calls `handle_payload()` method of all the unsolved responses in the `lst_required_responses`.

    Parameters
    ----------
    payload : dict
        The payload received from the server.
    """
    unsolved_responses = self.get_unsolved_responses()

    for response in unsolved_responses:
      response.handle_payload(payload)
    return

  def handle_notification(self, notification: dict) -> None:
    """
    This method is called when a notification is received from the server.
    This method calls `handle_notification()` method of all the unsolved responses in the `lst_required_responses`.

    Parameters
    ----------
    notification : dict
        The notification received from the server.
    """
    unsolved_responses = self.get_unsolved_responses()

    for response in unsolved_responses:
      response.handle_notification(notification)
    return

  def handle_heartbeat(self, heartbeat) -> None:
    """
    This method is called when a heartbeat is received from the server.
    This method calls `handle_heartbeat()` method of all the unsolved responses in the `lst_required_responses`.

    Parameters
    ----------
    heartbeat : dict
        The heartbeat received from the server.
    """
    unsolved_responses = self.get_unsolved_responses()

    for response in unsolved_responses:
      response.handle_heartbeat(heartbeat)
    return

  def callback(self):
    """
    Calls the resolved_callback only if the transaction is solved.
    """
    if self.__is_solved:
      if self.resolved_callback:
        self.resolved_callback()
      self.__is_finished = True
    return



import json
from copy import deepcopy


from ratio1 import Logger, const
from ratio1.bc import DefaultBlockEngine

def get_raw_whitelist():
  with open("./_local_cache/authorized_addrs", "r") as f:
    lines = f.readlines()
  return lines

if __name__ == '__main__':
  
  l = Logger("ENC", base_folder=".", app_folder="_local_cache")
  eng = DefaultBlockEngine(
    log=l, 
    name="default",
  )
  l.P("Raw whitelist:\n{}".format("".join(get_raw_whitelist())), color='d')
  l.P("Whitelist:\n{}".format("\n".join([
    a + (("  " + b) if len(b)>0 else "") for a,b in zip(*eng.whitelist_with_names)
  ])))
  
  
  
  to_add = [
    "0xai_A-Bn9grkqH1GUMTZUqHNzpX5DA6PqducH9_JKAlBx6YL         nen-aid02",
    "0xai_Amfnbt3N-qg2-qGtywZIPQBTVlAnoADVRmSAsdDhlQ-6             nen-2",
    "0xai_AwxGtRVqRlUrUoZNvf827uOswFmCkFXXguRmkpnyJBhQ",    
  ]
  
  eng.add_address_to_allowed("0xai_AszI8IKn985-IdDULXPrgYS5lnwg9PDmB95VGMYyltf4 bleo_core")
  eng.add_address_to_allowed("0xai_AszI8IKn985-IdDULXPrgYS5lnwg9PDmB95VGMYyltf4")
  eng.add_address_to_allowed("0xai_AszI8IKn985-IdDULXPrgYS5lnwg9PDmB95VGMYyltf4 bleo_core")
  eng.add_address_to_allowed("0xai_adresafoartegreistatotal.................")
  eng.add_address_to_allowed(to_add)
  
  
  l.P("Raw whitelist:\n{}".format("".join(get_raw_whitelist())), color='d')
  l.P("Whitelist:\n{}".format("\n".join([
    a + (("  " + b) if len(b)>0 else "") for a,b in zip(*eng.whitelist_with_names)
  ])))
